/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h> 
#include <netdb.h>

/* this function looks up host names and provides a list of 
   alternatives for addrs and aliases */ 

foreign_t
pl_gethostbyaddr(
  term_t hostaddr, 
  term_t cname, 
  term_t aliases, 
  term_t addrs, 
  foreign_t handle)
{
  struct in_addr in;
  term_t thing; 
  term_t list; 
  struct hostent *h; 		/* host entry to remember */ 
  char **p; 
  char *chars; 			/* characters in host name */ 
  ulong_t addr; 
  if (PL_is_atom(hostaddr)) { 
    list=PL_new_term_ref(); 
    thing=PL_new_term_ref(); 

    PL_get_atom_chars(hostaddr,&chars);  
    addr = inet_addr(chars);
    if ((int)addr!=-1) {
      h = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);

      /* check canonical name */ 
      if (!PL_unify_atom_chars(cname,h->h_name)) PL_fail; 

      /* check list of aliases */ 
      for (p=h->h_aliases; *p; p++); 
      PL_put_nil(list); 
      for (p--; p>=h->h_aliases; p--) {
	PL_put_atom_chars(thing,*p); 
	PL_cons_list(list,thing,list); 
      } 
      if (!PL_unify(aliases,list)) PL_fail; 

      /* check list of addresses */ 

      for (p=h->h_addr_list; *p; p++); 
      PL_put_nil(list); 
      for (p--; p>=h->h_addr_list; p--) { 
	(void) memcpy((char *)&in.s_addr, *p, 
		      sizeof (in.s_addr));
	chars= inet_ntoa(in); 
	PL_put_atom_chars(thing, chars); 
	PL_cons_list(list,thing,list); 
      }
      if (!PL_unify(addrs,list)) PL_fail; 
      PL_succeed; 
    } 
  } 
  PL_fail; 
}

install_t
pl_gethostbyaddr_install()
{ 
  PL_register_foreign("gethostbyaddr", 4, pl_gethostbyaddr, 0);
}
