/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
/* #include <strings.h> */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h> 

/* general password database iterator returns all matches
   for any specified pattern.  This is more expensive than simply
   matching UID and LOGIN, so a separate function geth is used to
   match them and return unique and simple results.  */

static struct hostent * hostent_copy(struct hostent *p) { 
  struct hostent *c = (struct hostent *) malloc (sizeof(struct hostent)); 
  if (p->h_name) { 
    c->h_name = (char *) malloc(strlen(p->h_name)+1); 
    strcpy(c->h_name,p->h_name); 
  } else { 
    c->h_name=NULL; 
  } 
  if (p->h_aliases) { 
    int alias,aliases; 
    for (aliases=0; p->h_aliases[aliases]; aliases++) ; 
    c->h_aliases = (char **) malloc((aliases+1)*sizeof(char *)); 
    for (alias=0; p->h_aliases[alias]; alias++) { 
      c->h_aliases[alias]=(char *) malloc(strlen(p->h_aliases[alias])+1); 
      strcpy(c->h_aliases[alias], p->h_aliases[alias]); 
    } 
    c->h_aliases[alias]=NULL; 
  } else { 
    c->h_aliases=NULL; 
  } 

  c->h_addrtype = p->h_addrtype; 
  c->h_length = p->h_length; 

  if (p->h_addr_list) { 
    int addr,addrs; 
    for (addrs=0; p->h_addr_list[addrs]; addrs++) ; 
    c->h_addr_list = (char **) malloc((addrs+1)*sizeof(char *)); 
    for (addr=0; p->h_addr_list[addr]; addr++) { 
      c->h_addr_list[addr]=(char *)malloc(c->h_length); 
      memcpy(c->h_addr_list[addr],p->h_addr_list[addr],p->h_length); 
    } 
    c->h_addr_list[addr]=NULL; 
  } else { 
    c->h_addr_list=NULL; 
  } 
  return c; 
} 

static void hostent_free (struct hostent *p) { 
   if (p->h_name) free(p->h_name); 
   if (p->h_aliases) { 
     int alias; 
     for (alias=0; p->h_aliases[alias]; alias++) free(p->h_aliases[alias]); 
     free(p->h_aliases); 
   } 
   if (p->h_addr_list) { 
     int addr; 
     for (addr=0; p->h_addr_list[addr]; addr++) free(p->h_addr_list[addr]); 
     free(p->h_addr_list); 
   } 
   free(p); 
} 

struct h_listelem { 
  struct h_listelem *next; 
  struct hostent *h; 
} ;

struct h_context { 
  struct h_listelem *head, *tail; 
} ; 

static void h_context_free(struct h_context *c) { 
  struct h_listelem *p; 
  while (c->head) { 
    p = c->head; 
    c->head = c->head->next; 
    hostent_free(p->h); free(p); 
  } 
} 

foreign_t
pl_gethostent(term_t name, 
      term_t aliases, 
      term_t addrs, 
      foreign_t handle)
{
  struct h_context *ctxt; 
  struct hostent *h; 
  struct h_listelem *p; 
  term_t my_addr_int; 
  term_t my_addr, my_addrs; 
  term_t my_alias, my_aliases; 
  char **q; 
  fid_t frame; 
  struct in_addr in; 
  char *name_chars = NULL; 
  if (PL_is_atom(name)) { 
    if (PL_get_atom_chars(name,&name_chars)) { 
      switch( PL_foreign_control(handle) )
      { case PL_FIRST_CALL:
	/* while this SEEMS like a good idea, it has the side 
           effect of only looking at the first match record; 
           this deletes the second value of the match record from 
           a secondary name service. This must be fixed! */ 

	  /* get the host entry */ 
	  h = gethostbyname(name_chars);
	  if (!h) PL_fail; 

	  /* create the alias list */ 
	  my_aliases = PL_new_term_ref();
	  my_alias = PL_new_term_ref();
	  /* lists are built tail-to-head so must start at end */
	  for (q=h->h_aliases; *q; q++);
	  PL_put_nil(my_aliases);
	  for (q--; q>=h->h_aliases; q--) {
	    PL_put_atom_chars(my_alias,*q);
	    PL_cons_list(my_aliases,my_alias,my_aliases);
	  }
	  if (!PL_unify(aliases,my_aliases)) PL_fail; 

	  /* create the host address list */ 
	  my_addrs = PL_new_term_ref();
	  my_addr = PL_new_term_ref();
          my_addr_int = PL_new_term_ref(); 
	  /* lists are built tail-to-head so must start at end */
	  for (q=h->h_addr_list; *q; q++);
	  PL_put_nil(my_addrs);
	  for (q--; q>=h->h_addr_list; q--) {
            char *chars; 
	    (void) memcpy((char *)&in.s_addr, *q, sizeof (in.s_addr));
	    chars= inet_ntoa(in);
	    PL_put_atom_chars(my_addr, chars);
	    PL_cons_list(my_addrs,my_addr,my_addrs);
	  }
	  if (!PL_unify(addrs,my_addrs)) PL_fail; 

	  PL_succeed;
	  break; 
        case PL_REDO:
        case PL_CUTTED: 
        default: 
	  PL_fail; 
      } 
    } else { 
      PL_fail; 
    } 
  } else if (PL_is_variable(name)) { 
    switch( PL_foreign_control(handle) )
    { case PL_FIRST_CALL:

	/* make a list of all matching entries, as of now */ 
	ctxt = (struct h_context *) malloc (sizeof(struct h_context)); 
	ctxt->head=ctxt->tail=NULL; 
	sethostent(0); 
	while ((h=gethostent())) { 
	  if ((!name_chars || strcmp(name_chars,h->h_name)==0)) { 
	    p = (struct h_listelem *) malloc(sizeof(struct h_listelem)); 
	    p->h = hostent_copy(h); 
	    if (!ctxt->head) ctxt->head=ctxt->tail=p; 
	    else { ctxt->tail->next = p;  ctxt->tail=p; } 
	    p->next = NULL; 
	  } 
	} 
        endhostent(); 

	/* now we have a linked list of matches; play them back
	   and simultaneously blow them away! */ 

	my_aliases = PL_new_term_ref();
	my_alias = PL_new_term_ref();
	my_addrs = PL_new_term_ref();
	my_addr = PL_new_term_ref();
        my_addr_int = PL_new_term_ref(); 
	while (ctxt->head) { 
	  frame = PL_open_foreign_frame(); 
	  h = ctxt->head->h;

	  /* create the alias list */ 
	  /* lists are built tail-to-head so must start at end */
	  for (q=h->h_aliases; *q; q++);
	  PL_put_nil(my_aliases);
	  for (q--; q>=h->h_aliases; q--) {
	    PL_put_atom_chars(my_alias,*q);
	    PL_cons_list(my_aliases,my_alias,my_aliases);
	  }

	  /* create the host address list */ 
	  /* lists are built tail-to-head so must start at end */
	  for (q=h->h_addr_list; *q; q++);
	  PL_put_nil(my_addrs);
	  for (q--; q>=h->h_addr_list; q--) {
	    char *chars; 
	    (void) memcpy((char *)&in.s_addr, *q, sizeof (in.s_addr));
	    chars= inet_ntoa(in);
	    PL_put_atom_chars(my_addr, chars);
	    PL_cons_list(my_addrs,my_addr,my_addrs);
	  }

	  if( PL_unify_atom_chars(name, h->h_name)
           && PL_unify(aliases, my_aliases) 
           && PL_unify(addrs, my_addrs)) {
	    p = ctxt->head; 
	    ctxt->head = ctxt->head->next; 
	    hostent_free(p->h); free(p); 
	    PL_close_foreign_frame(frame); 
	    PL_retry_address(ctxt);
	  } else { 
	    p = ctxt->head; 
	    ctxt->head = ctxt->head->next; 
	    hostent_free(p->h); free(p); 
	    PL_discard_foreign_frame(frame); 
          } 
        } 
	h_context_free(ctxt); 
	free(ctxt); 
	PL_fail; 

      case PL_REDO:

	/* next element is always the first element of the list */ 

	ctxt = PL_foreign_context_address(handle); 
	/* now we have a linked list of matches; play them back
	   and simultaneously blow them away! */ 

	my_aliases = PL_new_term_ref();
	my_alias = PL_new_term_ref();
	my_addrs = PL_new_term_ref();
	my_addr = PL_new_term_ref();
        my_addr_int = PL_new_term_ref(); 
	while (ctxt->head) { 
	  frame = PL_open_foreign_frame(); 
	  h = ctxt->head->h;

	  /* create the alias list */ 
	  /* lists are built tail-to-head so must start at end */
	  for (q=h->h_aliases; *q; q++);
	  PL_put_nil(my_aliases);
	  for (q--; q>=h->h_aliases; q--) {
	    PL_put_atom_chars(my_alias,*q);
	    PL_cons_list(my_aliases,my_alias,my_aliases);
	  }

	  /* create the host address list */ 
	  /* lists are built tail-to-head so must start at end */
	  for (q=h->h_addr_list; *q; q++);
	  PL_put_nil(my_addrs);
	  for (q--; q>=h->h_addr_list; q--) {
            char *chars; 
	    (void) memcpy((char *)&in.s_addr, *q, sizeof (in.s_addr));
	    chars= inet_ntoa(in);
	    PL_put_atom_chars(my_addr, chars);
	    PL_cons_list(my_addrs,my_addr,my_addrs);
	  }

	  if( PL_unify_atom_chars(name, h->h_name)
           && PL_unify(aliases, my_aliases) 
           && PL_unify(addrs, my_addrs)) {
	    p = ctxt->head; 
	    ctxt->head = ctxt->head->next; 
	    hostent_free(p->h); free(p); 
	    PL_close_foreign_frame(frame); 
	    PL_retry_address(ctxt);
	  } else { 
	    p = ctxt->head; 
	    ctxt->head = ctxt->head->next; 
	    hostent_free(p->h); free(p); 
	    PL_discard_foreign_frame(frame); 
          } 
        } 
	h_context_free(ctxt); 
	free(ctxt); 
	PL_fail; 

      case PL_CUTTED:

	/* cut context */ 
	ctxt = PL_foreign_context_address(handle); 
	h_context_free(ctxt); 
	free(ctxt); 
	PL_succeed;
    }
  } 
  PL_fail; 
} 

install_t
pl_gethostent_install()
{ 
  PL_register_foreign("gethostent", 3, pl_gethostent, PL_FA_NONDETERMINISTIC);
}
