/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

/* this function looks up host names and provides a list of 
   alternatives for addrs and aliases */ 

foreign_t
pl_getprotobynumber(
  term_t name, 
  term_t aliases,
  term_t proto, 
  foreign_t handle)
{
  struct protoent *h; 		/* protocol entry to remember */ 
  char **p; 
  int proto_val; 
  term_t thing = PL_new_term_ref(); 
  term_t list = PL_new_term_ref(); 
  if (PL_is_integer(proto)) { 
    PL_get_integer(proto,&proto_val);  
    h = getprotobynumber(proto_val); 
  } else { 
    PL_fail; 
  } 
  if (!h) PL_fail; 
  /* check protocol */ 
  if (!PL_unify_atom_chars(name,h->p_name)) PL_fail; 
  /* check list of aliases */ 
  for (p=h->p_aliases; *p; p++); 
  PL_put_nil(list); 
  for (p--; p>=h->p_aliases; p--) {
    PL_put_atom_chars(thing,*p); 
    PL_cons_list(list,thing,list); 
  } 
  if (!PL_unify(aliases,list)) PL_fail; 

  PL_succeed; 
}

install_t
pl_getprotobynumber_install()
{ 
  PL_register_foreign("getprotobynumber", 3, pl_getprotobynumber, 0);
}
