/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
/* #include <strings.h> */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h> 

/* general password database iterator returns all matches
   for any specified pattern.  This is more expensive than simply
   matching UID and LOGIN, so a separate function geth is used to
   match them and return unique and simple results.  */

static struct servent * servent_copy(struct servent *p) { 
  struct servent *c = (struct servent *) malloc (sizeof(struct servent)); 
  if (p->s_name) { 
    c->s_name = (char *) malloc(strlen(p->s_name)+1); 
    strcpy(c->s_name,p->s_name); 
  } else { 
    c->s_name=NULL; 
  } 
  if (p->s_aliases) { 
    int alias,aliases; 
    for (aliases=0; p->s_aliases[aliases]; aliases++) ; 
    c->s_aliases = (char **) malloc((aliases+1)*sizeof(char *)); 
    for (alias=0; p->s_aliases[alias]; alias++) { 
      c->s_aliases[alias]=(char *) malloc(strlen(p->s_aliases[alias])+1); 
      strcpy(c->s_aliases[alias], p->s_aliases[alias]); 
    } 
    c->s_aliases[alias]=NULL; 
  } else { 
    c->s_aliases=NULL; 
  } 

  c->s_port = p->s_port; 
  if (p->s_proto) { 
    c->s_proto = (char *) malloc(strlen(p->s_proto)+1); 
    strcpy(c->s_proto,p->s_proto); 
  } else { 
    c->s_proto=NULL; 
  } 

  return c; 
} 

static void servent_free (struct servent *p) { 
   if (p->s_name) free(p->s_name); 
   if (p->s_aliases) { 
     int alias; 
     for (alias=0; p->s_aliases[alias]; alias++) free(p->s_aliases[alias]); 
     free(p->s_aliases); 
   } 
   if (p->s_proto) free(p->s_proto); 
   free(p); 
} 

struct s_listelem { 
  struct s_listelem *next; 
  struct servent *h; 
} ;

struct s_context { 
  struct s_listelem *head, *tail; 
} ; 

static void s_context_free(struct s_context *c) { 
  struct s_listelem *p; 
  while (c->head) { 
    p = c->head; 
    c->head = c->head->next; 
    servent_free(p->h); free(p); 
  } 
} 

foreign_t
pl_getservent(term_t name, 
      term_t aliases, 
      term_t port, 
      term_t proto, 
      foreign_t handle)
{
  struct s_context *ctxt; 
  struct servent *h; 
  struct s_listelem *p; 
  term_t my_alias, my_aliases; 
  char **q; 
  fid_t frame; 
  char *name_chars = NULL; 
  if (PL_is_atom(name)) { 
    if (PL_get_atom_chars(name,&name_chars)) { 
      switch( PL_foreign_control(handle) )
      { case PL_FIRST_CALL:
	/* while this SEEMS like a good idea, it has the side 
           effect of only looking at the first match record; 
           this deletes the second value of the match record from 
           a secondary name service. This must be fixed! */ 

	  /* get the host entry */ 
	  if (PL_is_atom(proto)) { 
             char *proto_chars; 
             PL_get_atom_chars(proto,&proto_chars); 
             h = getservbyname(name_chars, proto_chars); 
          } else { 
	     h = getservbyname(name_chars,NULL);
          } 
	  if (!h) PL_fail; 

	  /* create the alias list */ 
	  my_aliases = PL_new_term_ref();
	  my_alias = PL_new_term_ref();
	  /* lists are built tail-to-head so must start at end */
	  for (q=h->s_aliases; *q; q++);
	  PL_put_nil(my_aliases);
	  for (q--; q>=h->s_aliases; q--) {
	    PL_put_atom_chars(my_alias,*q);
	    PL_cons_list(my_aliases,my_alias,my_aliases);
	  }
	  if (!PL_unify(aliases,my_aliases)) PL_fail; 
	  if (!PL_unify_integer(port,h->s_port)) PL_fail; 
	  if (!PL_unify_atom_chars(proto,h->s_proto)) PL_fail; 
	  PL_succeed;
	  break; 
        case PL_REDO:
        case PL_CUTTED: 
        default: 
	  PL_fail; 
      } 
    } else { 
      PL_fail; 
    } 
  } else if (PL_is_variable(name)) { 
    switch( PL_foreign_control(handle) )
    { case PL_FIRST_CALL:

	/* make a list of all matching entries, as of now */ 
	ctxt = (struct s_context *) malloc (sizeof(struct s_context)); 
	ctxt->head=ctxt->tail=NULL; 
	setservent(0); 
	while ((h=getservent())) { 
	  if ((!name_chars || strcmp(name_chars,h->s_name)==0)) { 
	    p = (struct s_listelem *) malloc(sizeof(struct s_listelem)); 
	    p->h = servent_copy(h); 
	    if (!ctxt->head) ctxt->head=ctxt->tail=p; 
	    else { ctxt->tail->next = p;  ctxt->tail=p; } 
	    p->next = NULL; 
	  } 
	} 
        endservent(); 

	/* now we have a linked list of matches; play them back
	   and simultaneously blow them away! */ 

	my_aliases = PL_new_term_ref();
	my_alias = PL_new_term_ref();
	while (ctxt->head) { 
	  frame = PL_open_foreign_frame(); 
	  h = ctxt->head->h;

	  /* create the alias list */ 
	  /* lists are built tail-to-head so must start at end */
	  for (q=h->s_aliases; *q; q++);
	  PL_put_nil(my_aliases);
	  for (q--; q>=h->s_aliases; q--) {
	    PL_put_atom_chars(my_alias,*q);
	    PL_cons_list(my_aliases,my_alias,my_aliases);
	  }

	  if( PL_unify_atom_chars(name, h->s_name)
           && PL_unify(aliases, my_aliases) 
           && PL_unify_integer(port, h->s_port)
           && PL_unify_atom_chars(proto, h->s_proto)) {
	    p = ctxt->head; 
	    ctxt->head = ctxt->head->next; 
	    servent_free(p->h); free(p); 
	    PL_close_foreign_frame(frame); 
	    PL_retry_address(ctxt);
	  } else { 
	    p = ctxt->head; 
	    ctxt->head = ctxt->head->next; 
	    servent_free(p->h); free(p); 
	    PL_discard_foreign_frame(frame); 
          } 
        } 
	s_context_free(ctxt); 
	free(ctxt); 
	PL_fail; 

      case PL_REDO:

	/* next element is always the first element of the list */ 

	ctxt = PL_foreign_context_address(handle); 
	/* now we have a linked list of matches; play them back
	   and simultaneously blow them away! */ 

	my_aliases = PL_new_term_ref();
	my_alias = PL_new_term_ref();
	while (ctxt->head) { 
	  frame = PL_open_foreign_frame(); 
	  h = ctxt->head->h;

	  /* create the alias list */ 
	  /* lists are built tail-to-head so must start at end */
	  for (q=h->s_aliases; *q; q++);
	  PL_put_nil(my_aliases);
	  for (q--; q>=h->s_aliases; q--) {
	    PL_put_atom_chars(my_alias,*q);
	    PL_cons_list(my_aliases,my_alias,my_aliases);
	  }

	  if( PL_unify_atom_chars(name, h->s_name)
           && PL_unify(aliases, my_aliases) 
           && PL_unify_integer(port, h->s_port)
           && PL_unify_atom_chars(proto, h->s_proto)) {
	    p = ctxt->head; 
	    ctxt->head = ctxt->head->next; 
	    servent_free(p->h); free(p); 
	    PL_close_foreign_frame(frame); 
	    PL_retry_address(ctxt);
	  } else { 
	    p = ctxt->head; 
	    ctxt->head = ctxt->head->next; 
	    servent_free(p->h); free(p); 
	    PL_discard_foreign_frame(frame); 
          } 
        } 
	s_context_free(ctxt); 
	free(ctxt); 
	PL_fail; 

      case PL_CUTTED:

	/* cut context */ 
	ctxt = PL_foreign_context_address(handle); 
	s_context_free(ctxt); 
	free(ctxt); 
	PL_succeed;
    }
  } 
  PL_fail; 
} 

install_t
pl_getservent_install()
{ 
  PL_register_foreign("getservent", 4, pl_getservent, PL_FA_NONDETERMINISTIC);
}
