/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <grp.h> 

/* general group database iterator goes through all groups
   looking for any specified pattern.  This is more expensive than
   simply matching GID or GROUP, so a separate function getgr is used
   to match them and return unique and simple results.  */

static struct group * group_copy(struct group *p) { 
  struct group *c = (struct group *) malloc (sizeof(struct group)); 
  char **q,**r; int count; 
  if (p->gr_name) { 
    c->gr_name = (char *) malloc(strlen(p->gr_name)+1); 
    strcpy(c->gr_name,p->gr_name); 
  } else { 
    c->gr_name=NULL; 
  } 
  if (p->gr_passwd) { 
    c->gr_passwd = (char *) malloc(strlen(p->gr_passwd)+1); 
    strcpy(c->gr_passwd,p->gr_passwd); 
  } else { 
    c->gr_passwd=NULL; 
  } 
  c->gr_gid = p->gr_gid; 
  for (q=p->gr_mem, count=0; *q; q++,count++) ; count++; 
  c->gr_mem = (char **) malloc(count*sizeof(char *)); 
  for (q=p->gr_mem, r=c->gr_mem; *q; q++,r++) { 
    *r=(char *)malloc(strlen(*q)+1); 
    strcpy(*r,*q); 
  } 
  *r=NULL; 
  return c; 
} 

static void group_free (struct group *p) { 
   char **q; 
   if (p->gr_name) free(p->gr_name); 
   if (p->gr_passwd) free(p->gr_passwd); 
   for (q=p->gr_mem; *q; q++) free(*q); 
   free(p); 
} 

struct gr_listelem { 
  struct gr_listelem *next; 
  struct group *gr; 
} ;

struct gr_context { 
  struct gr_listelem *head, *tail; 
} ; 

static void gr_context_free(struct gr_context *c) { 
  struct gr_listelem *p; 
  while (c->head) { 
    p = c->head; 
    c->head = c->head->next; 
    group_free(p->gr); free(p); 
  } 
} 

foreign_t
pl_group(
      term_t name,
      term_t passwd,
      term_t gid, 
      term_t members, 
      foreign_t handle)
{
  struct gr_context *ctxt; 
  struct group *gr; 
  struct gr_listelem *p; 
  char *name_chars = NULL; 
  int gid_val = -1; 
  term_t mem; 
  term_t mems; 
  term_t elem; 
  char **q; 
  if (PL_is_atom(name)) {
    if (PL_get_atom_chars(name,&name_chars)) {
      switch( PL_foreign_control(handle) )
      { case PL_FIRST_CALL:
        /* while this SEEMS like a good idea, it has the side
           effect of only looking at the first match record;
           this deletes the second value of the match record from
           a secondary name service. This must be fixed! */
          gr = getgrnam(name_chars);
          if (gr) { 
	    mems = PL_new_term_ref();
	    mem = PL_new_term_ref();
	    /* lists are built tail-to-head so must start at end */
	    for (q=gr->gr_mem; *q; q++);
	    PL_put_nil(mems);
	    for (q--; q>=gr->gr_mem; q--) {
	      PL_put_atom_chars(mem,*q);
	      PL_cons_list(mems,mem,mems);
	    }

	    if (PL_unify_integer(gid,gr->gr_gid)
	     && PL_unify_atom_chars(passwd,gr->gr_passwd)
	     && PL_unify(mems, members))
	      PL_succeed; 
            else 
	      PL_fail; 
          } else { 
            PL_fail;
          } 
          break;
        case PL_REDO:
        case PL_CUTTED:
        default:
          PL_fail;
      }
    } else {
      PL_fail;
    }
  } else if (PL_is_integer(gid)) {
    if (PL_get_integer(gid, (int *)&gid_val)) {
      switch( PL_foreign_control(handle) )
      { case PL_FIRST_CALL:
        /* while this SEEMS like a good idea, it has the side
           effect of only looking at the first match record;
           this deletes the second value of the match record from
           a secondary name service. This must be fixed! */
          gr = getgrgid(gid_val);
          if (gr) { 
	    mems = PL_new_term_ref();
	    mem = PL_new_term_ref();
	    /* lists are built tail-to-head so must start at end */
	    for (q=gr->gr_mem; *q; q++);
	    PL_put_nil(mems);
	    for (q--; q>=gr->gr_mem; q--) {
	      PL_put_atom_chars(mem,*q);
	      PL_cons_list(mems,mem,mems);
	    }
	    if (PL_unify_atom_chars(name, gr->gr_name)
	     && PL_unify_atom_chars(passwd,gr->gr_passwd)
	     && PL_unify(mems, members))
	      PL_succeed;
	    else
	      PL_fail;
          } else { 
            PL_fail;
	  } 
          break;
        case PL_REDO:
        case PL_CUTTED:
        default:
          PL_fail;
      }
    } else { 
      PL_fail; 
    } 
  } else { 
    switch( PL_foreign_control(handle) )
    { case PL_FIRST_CALL:

	/* first, avoid iteration if all is already lost */ 
	if (!PL_is_variable(name) 
	 && !PL_get_atom_chars(name,&name_chars))
	  PL_fail; 
	if (!PL_is_variable(gid) 
	 && !PL_get_integer(gid,&gid_val))
	  PL_fail; 
	if (!PL_is_variable(members) 
	 && !PL_is_list(members))
	  PL_fail; 

	/* make a list of all matching entries, as of now */ 

	ctxt = (struct gr_context *) malloc (sizeof(struct gr_context)); 
	ctxt->head=ctxt->tail=NULL; 
	setgrent(); 
	while ((gr=getgrent())) { 
	  if ((!name_chars || strcmp(name_chars,gr->gr_name)==0) 
	   && (gid_val==-1 || gid_val==gr->gr_gid)) {
	    p = (struct gr_listelem *) malloc(sizeof(struct gr_listelem)); 
	    p->gr = group_copy(gr); 
	    if (!ctxt->head) ctxt->head=ctxt->tail=p; 
	    else { ctxt->tail->next = p;  ctxt->tail=p; } 
	    p->next = NULL; 
	  } 
	} 

	/* now we have a linked list of matches; play them back
	   and simultaneously blow them away! The following code
	   is very subtle. We don't want to unify with a 
	   particular instance unless everything unifies. 
	   So we create foreign frames we can discard. */ 

	/* now, try to find a corresponding match from your list. 
	   stop on the first one! */ 
	elem=PL_new_term_ref(); 
	mems=PL_new_term_ref(); 
	while (ctxt->head) { 
	  fid_t frame = PL_open_foreign_frame(); 
	  gr = ctxt->head->gr;
	  PL_put_nil(mems); 
	  for (q=gr->gr_mem; *q; q++) ; 
	  for (q--; q>=gr->gr_mem; q--) { 
	      PL_put_atom_chars(elem,*q); 
	      PL_cons_list(mems,elem,mems); 
	  } 
	  if (PL_unify_atom_chars(name,gr->gr_name)
	   && PL_unify_atom_chars(passwd,gr->gr_passwd)
	   && PL_unify_integer(gid,gr->gr_gid)
	   && PL_unify(members,mems)) { 
	     p = ctxt->head; 
	     ctxt->head = ctxt->head->next; 
	     group_free(p->gr); free(p); 
	     PL_close_foreign_frame(frame); 
	     PL_retry_address(ctxt);
	  }  else { 
	     p = ctxt->head; 
	     ctxt->head = ctxt->head->next; 
	     group_free(p->gr); free(p); 
	     PL_discard_foreign_frame(frame); 
	  } 
	} 
	gr_context_free(ctxt); 
	free(ctxt); 
	PL_fail; 

      case PL_REDO:

	/* next element is always the first element of the list */ 

	ctxt = PL_foreign_context_address(handle); 

	/* now, try to find a corresponding match from your list. 
	   stop on the first one! */ 
	elem=PL_new_term_ref(); 
	mems=PL_new_term_ref(); 
	while (ctxt->head) { 
	  fid_t frame = PL_open_foreign_frame(); 
	  gr = ctxt->head->gr;
	  PL_put_nil(mems); 
	  for (q=gr->gr_mem; *q; q++) ; 
	  for (q--; q>=gr->gr_mem; q--) { 
	      PL_put_atom_chars(elem,*q); 
	      PL_cons_list(mems,elem,mems); 
	  } 
	  if (PL_unify_atom_chars(name,gr->gr_name)
	   && PL_unify_atom_chars(passwd,gr->gr_passwd)
	   && PL_unify_integer(gid,gr->gr_gid)
	   && PL_unify(members,mems)) { 
	     p = ctxt->head; 
	     ctxt->head = ctxt->head->next; 
	     group_free(p->gr); free(p); 
	     PL_close_foreign_frame(frame); 
	     PL_retry_address(ctxt);
	  }  else { 
	     p = ctxt->head; 
	     ctxt->head = ctxt->head->next; 
	     group_free(p->gr); free(p); 
	     PL_discard_foreign_frame(frame); 
	  } 
	} 
	gr_context_free(ctxt); 
	free(ctxt); 
	PL_fail; 

      case PL_CUTTED:

	/* cut context */ 
	ctxt = PL_foreign_context_address(handle); 
	gr_context_free(ctxt); 
	free(ctxt); 
	PL_succeed;
    }
  } 
  PL_fail; 
} 

install_t
pl_group_install()
{ 
  PL_register_foreign("group", 4, pl_group, PL_FA_NONDETERMINISTIC);
}
