/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h> 

/* change the mode of a file to match a template, if necessary */ 
foreign_t 
pl_path_mode(term_t path, term_t mode) 
{ 
  char *path_val; 
  int mode_val; 
  if (!PL_get_atom_chars(path,&path_val)) PL_fail; 
  if (PL_get_integer(mode,&mode_val)) { 
    struct stat buf; 
    if (stat(path_val,&buf)!=0) PL_fail; 
    if ((buf.st_mode&07777) != (mode_val&07777)) { 
      if (!chmod(path_val, mode_val&07777)) PL_succeed; 
      else PL_fail; 
    } else PL_succeed; 
  } else if (PL_is_variable(mode)) { 
    struct stat buf; 
    if (stat(path_val,&buf)!=0) PL_fail; 
    if (PL_unify_integer(mode,buf.st_mode&07777)) PL_succeed; 
    else PL_fail; 
  } else PL_fail; 
} 

foreign_t
pl_mode_ascii(term_t mode, term_t ascii) { 
  int mode_val; 
  char *ascii_val; 
  char buf[10]; 
  int i; 
  if (PL_get_integer(mode, &mode_val)) { 
    buf[9]='\0'; 
    for (i=0; i<9; i++) buf[i]='-'; 
    if (mode_val&0400) buf[0]='r'; 
    if (mode_val&0200) buf[1]='w'; 
    if ((mode_val&04100)==04100) buf[2]='s'; 
    else if (mode_val&04000) buf[2]='S'; 
    else if (mode_val&0100) buf[2]='x'; 
    if (mode_val&0040) buf[3]='r'; 
    if (mode_val&0020) buf[4]='w'; 
    if ((mode_val&02010)==02010) buf[5]='s'; 
    else if (mode_val&02000) buf[5]='S'; 
    else if (mode_val&0010) buf[5]='x'; 
    if (mode_val&0004) buf[6]='r'; 
    if (mode_val&0002) buf[7]='w'; 
    if ((mode_val&01001)==01001) buf[8]='t'; 
    else if (mode_val&01000) buf[8]='T'; 
    else if (mode_val&0001) buf[8]='x'; 
    if (PL_unify_atom_chars(ascii, buf)) PL_succeed; 
    else PL_fail; 
  } else if (PL_get_atom_chars(ascii, &ascii_val)) { 
    if (strlen(ascii_val)<9) PL_fail; 
    mode_val=0; 
    if (ascii_val[0]=='r') mode_val |= 0400; 
    else if (ascii_val[0]!='-') PL_fail; 
    if (ascii_val[1]=='w') mode_val |= 0200; 
    else if (ascii_val[1]!='-') PL_fail; 
    if (ascii_val[1]=='x') mode_val |= 0100; 
    else if (ascii_val[2]=='s') mode_val |= 04100; 
    else if (ascii_val[2]=='S') mode_val |= 04000; 
    else if (ascii_val[2]!='-') PL_fail; 
    if (ascii_val[3]=='r') mode_val |= 040; 
    else if (ascii_val[3]!='-') PL_fail; 
    if (ascii_val[4]=='w') mode_val |= 020; 
    else if (ascii_val[4]!='-') PL_fail; 
    if (ascii_val[5]=='x') mode_val |= 010; 
    else if (ascii_val[5]=='s') mode_val |= 02010; 
    else if (ascii_val[5]=='S') mode_val |= 02000; 
    else if (ascii_val[5]!='-') PL_fail; 
    if (ascii_val[6]=='r') mode_val |= 04; 
    else if (ascii_val[6]!='-') PL_fail; 
    if (ascii_val[7]=='w') mode_val |= 02; 
    else if (ascii_val[7]!='-') PL_fail; 
    if (ascii_val[8]=='x') mode_val |= 01; 
    else if (ascii_val[8]=='t') mode_val |= 01001; 
    else if (ascii_val[8]=='T') mode_val |= 01000; 
    else if (ascii_val[8]!='-') PL_fail; 
    if (PL_unify_integer(mode,mode_val)) PL_succeed; 
    else PL_fail; 
  } 
  PL_fail; 
} 

foreign_t 
pl_path_mask(term_t path, term_t mask) { 
  char *path_val, *mask_val, *p; 
  int to_add, to_remove, types, newmode; 
  enum {NONE, ADD, REMOVE} state; 
  struct stat buf; 
  if (!PL_get_atom_chars(path,&path_val)) PL_fail; 
  if (!PL_get_atom_chars(mask,&mask_val)) PL_fail; 
  /* construct to_add and to_remove masks */ 
  to_add=to_remove=0; types=0; state=NONE; 
  p=mask_val; 
  while (*p) { 
    switch (*p) { 
      case 'u': 
	types |= 4; 
        break; 
      case 'g': 
	types |= 2; 
        break; 
      case 'o': 
	types |= 1; 
        break; 
      case 'r': 
	switch (state) { 
	  case ADD: 
	    if (types&4) { to_add|= 4<<6; to_remove &=~(4<<6); } 
	    if (types&2) { to_add|= 4<<3; to_remove &=~(4<<3); } 
	    if (types&1) { to_add|= 4<<0; to_remove &=~(4<<0); } 
	    break; 
	  case REMOVE: 
	    if (types&4) { to_remove|= 4<<6; to_add &=~(4<<6); } 
	    if (types&2) { to_remove|= 4<<3; to_add &=~(4<<3); } 
	    if (types&1) { to_remove|= 4<<0; to_add &=~(4<<0); } 
	    break; 
          case NONE: 
	   fprintf(stderr,"illegal mode mask!\n"); 
	   PL_fail; 
        } 
        break; 
      case 'w': 
	switch (state) { 
	  case ADD: 
	    if (types&4) { to_add|= 2<<6; to_remove &=~(2<<6); } 
	    if (types&2) { to_add|= 2<<3; to_remove &=~(2<<3); } 
	    if (types&1) { to_add|= 2<<0; to_remove &=~(2<<0); } 
	    break; 
	  case REMOVE: 
	    if (types&4) { to_remove|= 2<<6; to_add &=~(2<<6); } 
	    if (types&2) { to_remove|= 2<<3; to_add &=~(2<<3); } 
	    if (types&1) { to_remove|= 2<<0; to_add &=~(2<<0); } 
	    break; 
          case NONE: 
	   fprintf(stderr,"illegal mode mask!\n"); 
	   PL_fail; 
        } 
        break; 
      case 'x': 
	switch (state) { 
	  case ADD: 
	    if (types&4) { to_add|= 1<<6; to_remove &=~(1<<6); } 
	    if (types&2) { to_add|= 1<<3; to_remove &=~(1<<3); } 
	    if (types&1) { to_add|= 1<<0; to_remove &=~(1<<0); } 
	    break; 
	  case REMOVE: 
	    if (types&4) { to_remove|= 1<<6; to_add &=~(1<<6); } 
	    if (types&2) { to_remove|= 1<<3; to_add &=~(1<<3); } 
	    if (types&1) { to_remove|= 1<<0; to_add &=~(1<<0); } 
	    break; 
          case NONE: 
	   fprintf(stderr,"illegal mode mask!\n"); 
	   PL_fail; 
        } 
        break; 
      case 's': 
	switch (state) { 
	  case ADD: 
            if (types&4) { to_add|= 04000; to_remove&=~04000; } 
            if (types&2) { to_add|= 02000; to_remove&=~02000; } 
            if (types&1) { to_add|= 01000; to_remove&=~01000; } 
	    break; 
	  case REMOVE: 
            if (types&4) { to_remove|= 04000; to_add&=~04000; } 
            if (types&2) { to_remove|= 02000; to_add&=~02000; } 
            if (types&1) { to_remove|= 01000; to_add&=~01000; } 
	    break; 
          case NONE: 
	   fprintf(stderr,"illegal mode mask!\n"); 
	   PL_fail; 
        } 
        break; 
      case '+': 
        state=ADD; 
        break; 
      case '-': 
        state=REMOVE; 
        break; 
      case ',': 
      case ' ': 
      case '\t':
      default: 
	/* reset for next word */ 
        types=0; state=NONE; 
        break; 
    } 
    p++; 
  } 
  if (stat(path_val,&buf)!=0) PL_fail; 
  newmode = (buf.st_mode & ~to_remove) | to_add; 
  if ((buf.st_mode&07777) != (newmode&07777)) { 
    if (!chmod(path_val, newmode&07777)) PL_succeed; 
    else PL_fail; 
  } else PL_succeed; 
} 

install_t
pl_mode_install()
{ 
  PL_register_foreign("path_mode", 2, pl_path_mode, 0);
  PL_register_foreign("path_mask", 2, pl_path_mask, 0);
  PL_register_foreign("mode_ascii", 2, pl_mode_ascii, 0);
}

