/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h> 
#include <sys/wait.h> 
#include <signal.h> 

struct li_listelem {
  struct li_listelem *next;
  char *line; 
} ;

struct li_context {
  struct li_listelem *head, *tail;
} ;

/* This program executes a command and 
   returns its output one line at a time */ 

foreign_t 
pl_output_range(term_t command, term_t first, term_t last, term_t output, foreign_t handle) 
{ 
  int pid; 
  int stat; 
  int filedes[2]; 
  char *path_val; 
  char *command_val; 
  int first_val; 
  int last_val; 
  struct li_context *ctxt; 

  /* get parameters */ 
  if (!PL_get_atom_chars(command,&command_val)) PL_fail; 
  if (!PL_get_integer(first,&first_val)) PL_fail; 
  if (first_val<0) PL_fail; 
  if (!PL_get_integer(last,&last_val)) PL_fail; 
  if (last_val<0) PL_fail; 

  /* if this is the first call, then read command output */ 
  switch( PL_foreign_control(handle) )
  { case PL_FIRST_CALL:

      /* set up pipe and fork */ 
      if (pipe(filedes)) PL_fail; 
      if ((pid=fork())) { 	/* parent */ 
        char buf[1024]; 
	int count=0; 
	FILE *in=fdopen(filedes[0],"r"); 
	close(filedes[1]); /* close stdout so that child close exits wait */ 
	ctxt = (struct li_context *) malloc(sizeof(struct li_context)); 
        ctxt->head = ctxt->tail = NULL; 
	/* read from stdout of child */ 
	/* consume initial header */ 
        while (count<first_val && !feof(in) && fgets(buf,1024,in)) count++; 
	/* now read lines */ 
	while (!feof(in) && count<last_val) { 
	  int len; 
 	  struct li_listelem *l; 
	  if (fgets(buf,1024,in)) { 
	    len=strlen(buf); 
	    if (buf[len-1]=='\n') buf[len-1]='\0'; 
  /* fprintf(stderr,"got %s\n",buf); */
	    l = (struct li_listelem *) malloc(sizeof(struct li_listelem)); 
	    l->line = (char *) malloc (strlen(buf)+1); 
	    strcpy(l->line,buf); 
	    if (!ctxt->head) ctxt->head=ctxt->tail=l;
	    else { ctxt->tail->next = l;  ctxt->tail=l; }
	    l->next = NULL;
	  } 
	} 
	/* close input side of pipe */ 
	fclose(in); 
        kill(pid,SIGQUIT); 
	waitpid(pid,&stat,0);
	if (stat!=0) PL_fail; 
      } else { 	  	/* child */ 
	char *args[128]; 
	char buf[1024], *p, *q; 
	int arg = 0; 
	int end = 0; 
	close(filedes[0]); 	/* force pipe close on eof from parent */ 
	close(1); 		/* dup input pipe to standard input */ 
	if (dup(filedes[1])!=1) { 
	  fprintf(stderr,"can't dup standard output!\n"); 
	  exit(1); 
	} 
	close(0); 		/* will not send input */ 

	strcpy(buf,command_val); 
	p=buf; 

	/* parse out command path */ 
	while (*p==' ' || *p=='\t' || *p=='\n') p++; 
	q=p; 
	while (*q!=' ' && *q!='\t' && *q!='\n' && *q!='\0') q++; 
	if (!*q) end=1; 
	*q=0; 
	path_val=p; args[arg++]=p; 
	p=q+1; 

	/* parse out command arguments */ 
	while (!end) { 
	  while (*p==' ' || *p=='\t' || *p=='\n') p++; 
	  q=p; 
	  while (*q!=' ' && *q!='\t' && *q!='\n' && *q!='\0') q++; 
	  if (!*q) end=1; 
	  *q=0; 
	  args[arg++]=p; 
	  p=q+1; 
	} 
	args[arg++]=NULL; 
	execv(path_val, args); 
	fprintf(stderr,"exec failed!\n");   
	PL_fail; 
      } 
      while (ctxt->head) {
	fid_t frame = PL_open_foreign_frame();
	if (PL_unify_atom_chars(output,ctxt->head->line)) {
	   struct li_listelem *p = ctxt->head;
	   ctxt->head = ctxt->head->next;
	   free(p->line); free(p); 
	   PL_close_foreign_frame(frame);
	   PL_retry_address(ctxt);
	}  else {
	   struct li_listelem *p = ctxt->head;
	   ctxt->head = ctxt->head->next;
	   free(p->line); free(p); 
	   PL_discard_foreign_frame(frame);
	}
      }
      PL_fail; 
      break; 
    case PL_REDO:
      /* next element is always the first element of the list */
      ctxt = PL_foreign_context_address(handle);
      while (ctxt->head) {
	fid_t frame = PL_open_foreign_frame();
	if (PL_unify_atom_chars(output,ctxt->head->line)) {
	   struct li_listelem *p = ctxt->head;
	   ctxt->head = ctxt->head->next;
	   free(p->line); free(p); 
	   PL_close_foreign_frame(frame);
	   PL_retry_address(ctxt);
	}  else {
	   struct li_listelem *p = ctxt->head;
	   ctxt->head = ctxt->head->next;
	   free(p->line); free(p); 
	   PL_discard_foreign_frame(frame);
	}
      }
      PL_fail; 
      break; 
    case PL_CUTTED: 
      while (ctxt->head) { 
        struct li_listelem *p = ctxt->head;
        ctxt->head = ctxt->head->next;
        free(p->line); free(p); 
      } 
      PL_fail; 
      break; 
  } 
  PL_succeed; 
} 

install_t
pl_output_range_install()
{ 
  PL_register_foreign("output_range", 4, pl_output_range, PL_FA_NONDETERMINISTIC);
}
