/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h> 

/* see man regexp */ 
static char pat[256]; 
static char *patp; 
#define INIT
#define GETC() (*(patp++))
#define UNGETC(c) (patp--)
#define PEEKC() (*patp)
#define RETURN(ptr) return (ptr)
#define ERROR(val) return 0
#include <regexp.h> 

/* this function does regexp matching on an atom */ 

foreign_t 
pl_match(term_t line, term_t pattern, term_t match, foreign_t handle) 
{ 
  char *line_val; 
  char *pattern_val; 
  char matched[256]; 
  char buf[10240]; 

  /* get parameters */ 
  if (!PL_get_atom_chars(line,&line_val)) PL_fail; 
  if (!PL_get_atom_chars(pattern,&pattern_val)) PL_fail; 
  
  strcpy(pat,pattern_val); patp=pat; 
  if (!compile(line_val,buf,buf+10240-1,'\0')) PL_fail; 
  if (!step(line_val,buf)) PL_fail; 
   
  strncpy(matched,loc1,loc2-loc1); 
  *(matched+(loc2-loc1))='\0'; 
  if (PL_unify_atom_chars(match,matched)) PL_succeed; 
  PL_fail; 
} 


/* this function splits an atom */ 

foreign_t 
pl_split(term_t line, term_t pattern, term_t list, foreign_t handle) 
{ 
  char *line_val; 
  char *pattern_val; 
  char matched[256]; 
  char *linep; 
  term_t my_list; 
  term_t my_list_elem; 
  char *begin[256], *end[256]; 
  int where; 
  char buf[10240]; 

  /* get parameters */ 
  if (!PL_get_atom_chars(line,&line_val)) PL_fail; 
  if (!PL_get_atom_chars(pattern,&pattern_val)) PL_fail; 
  
  strcpy(pat,pattern_val); patp=pat; 
  if (!compile(line_val,buf,buf+10240-1,'\0')) PL_fail; 
  
  /* split at all match points */ 
  where=0; 
  linep=line_val; 
  if (step(linep,buf)) { 
    if (loc1==linep) { 
      begin[where]=linep; 
      end[where]=linep; 
    } else { 
      begin[where]=linep; 
      end[where]=loc1; 
    } 
/* fprintf(stderr,"where=%d\n",where);  */
/* fprintf(stderr," begin=%s\n", begin[where]);  */
/* fprintf(stderr," end=%s\n", end[where]);  */
    where++; 
    linep=loc2; 
    while (step(linep,buf)) { 
      begin[where]=linep; 
      end[where]=loc1; 
/* fprintf(stderr,"where=%d\n",where);  */
/* fprintf(stderr," begin=%s\n", begin[where]);  */
/* fprintf(stderr," end=%s\n", end[where]);  */
      where++; 
      linep=loc2; 
    } 
    begin[where]=linep; 
    end[where]=linep+strlen(linep); 
/* fprintf(stderr,"where=%d\n",where);  */
/* fprintf(stderr," begin=%s\n", begin[where]);  */
/* fprintf(stderr," end=%s\n", end[where]);  */
    where++; 
  } else { 
    begin[where]=linep; 
    end[where]=linep+strlen(linep); 
/* fprintf(stderr,"where=%d\n",where);  */
/* fprintf(stderr," begin=%s\n", begin[where]);  */
/* fprintf(stderr," end=%s\n", end[where]);  */
    where++; 
  } 

  /* place this into a prolog list backward */ 
  my_list=PL_new_term_ref(); 
  my_list_elem=PL_new_term_ref(); 
  PL_put_nil(my_list); 
/* fprintf(stderr,"constructing list now\n");  */
  while (where>0) { 
    where--; 
/* fprintf(stderr,"where=%d\n",where);  */
/* fprintf(stderr," begin=%s\n", begin[where]);  */
/* fprintf(stderr," end=%s\n", end[where]);  */
    strncpy(matched,begin[where],end[where]-begin[where]); 
    *(matched+(end[where]-begin[where])) = '\0'; 
    PL_put_atom_chars(my_list_elem,matched); 
    PL_cons_list(my_list,my_list_elem,my_list);
  } 
   
  if (PL_unify(list,my_list)) PL_succeed; 
  PL_fail; 
} 

/* this function does an unlimited replace of a regexp */ 

foreign_t 
pl_replace(term_t line, term_t pattern, term_t replacement, 
     term_t result, foreign_t handle) 
{ 
  char *line_val; 
  char line_in[1024]; 
  char line_out[1024]; 
  char *pattern_val; 
  char *replacement_val; 
  char buf[10240]; 
  char *p, *q; 

  /* get parameters */ 
  if (!PL_get_atom_chars(line,&line_val)) PL_fail; 
  if (!PL_get_atom_chars(pattern,&pattern_val)) PL_fail; 
  if (!PL_get_atom_chars(replacement,&replacement_val)) PL_fail; 
  
  strcpy(pat,pattern_val); patp=pat; 
  if (!compile(line_val,buf,buf+10240-1,'\0')) PL_fail; 
  strcpy(line_in,line_val); 
  
  while (step(line_in,buf)) { 
    for (p=line_in, q=line_out; p<loc1; p++,q++) *q=*p; 
    for (p=replacement_val; *p; p++, q++) *q=*p; 
    for (p=loc2; *p; p++, q++) *q=*p; 
    *q='\0'; 
    for (p=line_out, q=line_in; *p; p++,q++) *q=*p; 
    *q='\0'; 
/* fprintf(stderr,"result is %s\n", line_in);  */
  } 

  if (PL_unify_atom_chars(result,line_in)) PL_succeed; 
  PL_fail; 
} 

/* functions for dealing with tabular files must split 
   lines internally using regexps, so must be filed here. */ 

/* assert the content of a single row in a tabular file, with an index */ 
foreign_t
pl_table_assert(term_t table, term_t content, term_t index, term_t delim) { 
  char *table_val, *content_val, *delim_val; int index_val; 
  char dirname[256]; 
  char filename[256]; 
  struct stat tstat, dstat, nstat; 
  char *begin[256], *end[256]; 	/* parser table */ 
  int where; 			/* where in table are we */ 
  char patbuf[10240]; 		/* pattern buffer */ 
  char linbuf[1024]; 		/* line buffer */ 
  int smode; 
  FILE *ffd; 
  if (!PL_get_atom_chars(table,&table_val)) PL_fail; 
  if (!PL_get_atom_chars(content,&content_val)) PL_fail; 
  if (!PL_get_atom_chars(delim,&delim_val)) PL_fail; 
  if (!PL_get_integer(index,&index_val)) PL_fail; 
  strcpy(pat,delim_val); patp=pat; 
  if (!compile(NULL,patbuf,patbuf+10240-1,'\0')) PL_fail; 
  if (stat(table_val, &tstat)) PL_fail; 
  strcpy(dirname,table_val); strcat(dirname,"+++"); 
  if (stat(dirname, &dstat)) { /* no such thing -> create */ 
    FILE *tfd; 
    char *linep, *fieldp; 
    if (mkdir(dirname, tstat.st_mode|0700)) PL_fail; 
    if (! (tfd=fopen(table_val, "r"))) { 
      rmdir(dirname); 
      PL_fail; 
    } 
    while (!feof(tfd) && fgets(linbuf, 1024, tfd)) { 
      /* split patbuf into fields, as in other parsers */ 
      where=0; 
      linep=linbuf; 
      if (step(linep,patbuf)) { 
	if (loc1==linep) { 
	  begin[where]=linep; 
	  end[where]=linep; 
	} else { 
	  begin[where]=linep; 
	  end[where]=loc1; 
	  *(end[where]) = '\0'; 
	} 
/* fprintf(stderr,"where=%d\n",where);  */
/* fprintf(stderr," begin=%s\n", begin[where]);  */
/* fprintf(stderr," end=%s\n", end[where]);  */
	where++; 
	linep=loc2; 
	while (step(linep,patbuf)) { 
	  begin[where]=linep; 
	  end[where]=loc1; 
	  *(end[where]) = '\0'; 
/* fprintf(stderr,"where=%d\n",where);  */
/* fprintf(stderr," begin=%s\n", begin[where]);  */
/* fprintf(stderr," end=%s\n", end[where]);  */
	  where++; 
	  linep=loc2; 
	} 
	begin[where]=linep; 
	end[where]=linep+strlen(linep); 
/* fprintf(stderr,"where=%d\n",where);  */
/* fprintf(stderr," begin=%s\n", begin[where]);  */
/* fprintf(stderr," end=%s\n", end[where]);  */
	where++; 
      } else { 
	begin[where]=linep; 
	end[where]=linep+strlen(linep); 
/* fprintf(stderr,"where=%d\n",where);  */
/* fprintf(stderr," begin=%s\n", begin[where]);  */
/* fprintf(stderr," end=%s\n", end[where]);  */
	where++; 
      } 
      /* have begin, end, and where initialized, and strings delimited */ 
      /* now select and translate a field into a file index, and 
         write the line to the file */ 
      fieldp = begin[index_val]; 
      if (fieldp != NULL) { 
	/* first transliterate evil characters */ 
	int i; 
        for (i=0; fieldp[i]; i++) { 
	  switch (fieldp[i]) { 
	    case ' ': 
	    case '/': 
	    case '|': 
	    case '~': 
	    case '*': 
	    case '!': 
	    case '%': 
	      fieldp[i]='_'; 
              break; 
	    default: ; 
	  } 
        } 
	/* then make a file with that name */ 
        strcpy(filename,dirname); 
	strcat(filename,"/"); 
	strcat(filename, fieldp); 
        if ((ffd=fopen(filename,"w"))!=0) { 
          fprintf(ffd,"%s",linbuf); 
	  fclose(ffd); 
	  /* change modes to match master file */ 
          chown (filename, tstat.st_uid, tstat.st_gid); 
          smode = tstat.st_mode &~(07111);
          chmod(filename,smode); 
        } 
      } else { 
        fprintf(stderr, "field index failed!\n"); 
      } 
    } 
  } 
  PL_succeed; 
} 

foreign_t
pl_table_retract(term_t table, term_t content, term_t index, term_t delim) { 
  char *table_val, *index_val, *content_val, *delim_val; 
  PL_succeed; 

} 


install_t
pl_regexp_install()
{ 
  PL_register_foreign("split", 3, pl_split, 0);
  PL_register_foreign("match", 3, pl_match, 0);
  PL_register_foreign("replace", 4, pl_replace, 0);
  PL_register_foreign("table_assert", 4, pl_table_assert, 0);
}
