/*********************************************************
## This file is part of Prodog.  All files in the distribution of Prodog
## are Copyright 2000 by
##
##     Alva L.  Couch,
##     Associate Professor of Computer Science,
##     Tufts University, Medford MA 02155
##     email: couch@eecs.tufts.edu
##     phone: 617-627-3674
##
## All rights reserved.
##
## Redistribution, modification, and use are permitted provided that this 
## entire copyright notice is duplicated in all such copies, and that any
## documentation, announcements, and other materials related to such
## distribution and use acknowledge that the software was developed by
## Alva L.  Couch at Tufts University in Medford, Massachusetts, USA.
## No charge, other than an "at-cost" distribution fee, may be charged
## for copies, derivations, or distributions of this material without
## the express written consent of the copyright holder.  Neither the
## name of the University nor the names of the authors may be used to
## endorse or promote products derived from this material without
## specific prior written permission.  THIS SOFTWARE IS PROVIDED ``AS
## IS'' AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
## WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
## FITNESS FOR ANY PARTICULAR PURPOSE.
 *********************************************************/
/*  Include file depends on local installation */
#include <SWI-Prolog.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h> 
#include <sys/stat.h> 

/* provide a file interface that reads file data */

foreign_t
pl_stat10(
  term_t path, 
  term_t mode, 
  term_t ino, 
  term_t nlink, 
  term_t uid, 
  term_t gid, 
  term_t size, 
  term_t atim, 
  term_t mtim, 
  term_t ctim, 
  foreign_t handle)
{
  struct stat buf; 
  char *chars; 
  char out[256]; 
  if (!PL_get_atom_chars(path,&chars))		PL_fail; 
  if (stat(chars,&buf)!=0) 			PL_fail; 
  sprintf(out,"0o%07o",(unsigned int)buf.st_mode); 
  if (!PL_unify_atom_chars(mode,out)) 		PL_fail; 
  if (!PL_unify_integer(ino,buf.st_ino)) 	PL_fail; 
  if (!PL_unify_integer(nlink,buf.st_nlink)) 	PL_fail; 
  if (!PL_unify_integer(uid,buf.st_uid)) 	PL_fail; 
  if (!PL_unify_integer(gid,buf.st_gid)) 	PL_fail; 
  if (!PL_unify_integer(size,buf.st_size)) 	PL_fail; 
  if (!PL_unify_integer(atim,buf.st_atime)) 	PL_fail; 
  if (!PL_unify_integer(mtim,buf.st_mtime)) 	PL_fail; 
  if (!PL_unify_integer(ctim,buf.st_ctime)) 	PL_fail; 
  PL_succeed;
}

foreign_t
pl_stat5(
  term_t path, 
  term_t dev, 
  term_t rdev, 
  term_t blksize, 
  term_t blocks, 
  foreign_t handle)
{
  struct stat buf; 
  char *chars; 
  if (!PL_get_atom_chars(path,&chars)) 		PL_fail; 
  if (stat(chars,&buf)!=0) 			PL_fail; 
  if (!PL_unify_integer(dev,buf.st_dev)) 	PL_fail; 
  if (!PL_unify_integer(rdev,buf.st_rdev)) 	PL_fail; 
  if (!PL_unify_integer(blksize,buf.st_blksize)) PL_fail; 
  if (!PL_unify_integer(blocks,buf.st_blocks))	PL_fail; 
  PL_succeed;
}

foreign_t
pl_lstat10(
  term_t path, 
  term_t mode, 
  term_t ino, 
  term_t nlink, 
  term_t uid, 
  term_t gid, 
  term_t size, 
  term_t atim, 
  term_t mtim, 
  term_t ctim, 
  foreign_t handle)
{
  struct stat buf; 
  char *chars; 
  char out[256]; 
  if (! PL_get_atom_chars(path,&chars)) 	PL_fail; 
  if (lstat(chars,&buf)!=0) 			PL_fail; 
  sprintf(out,"0o%07o",(unsigned int) buf.st_mode); 
  if (!PL_unify_atom_chars(mode,out)) 		PL_fail; 
  if (!PL_unify_integer(ino,buf.st_ino)) 	PL_fail; 
  if (!PL_unify_integer(nlink,buf.st_nlink))	PL_fail; 
  if (!PL_unify_integer(uid,buf.st_uid))	PL_fail; 
  if (!PL_unify_integer(gid,buf.st_gid))	PL_fail; 
  if (!PL_unify_integer(size,buf.st_size))	PL_fail; 
  if (!PL_unify_integer(atim,buf.st_atime))	PL_fail; 
  if (!PL_unify_integer(mtim,buf.st_mtime))	PL_fail; 
  if (!PL_unify_integer(ctim,buf.st_ctime))	PL_fail; 
  PL_succeed;
}

foreign_t
pl_lstat5(
  term_t path, 
  term_t dev, 
  term_t rdev, 
  term_t blksize, 
  term_t blocks, 
  foreign_t handle)
{
  struct stat buf; 
  char *chars; 
  if (! PL_get_atom_chars(path,&chars))		PL_fail; 
  if (lstat(chars,&buf)!=0)			PL_fail; 
  if (!PL_unify_integer(dev,buf.st_dev))	PL_fail; 
  if (!PL_unify_integer(rdev,buf.st_rdev))	PL_fail; 
  if (!PL_unify_integer(blksize,buf.st_blksize)) PL_fail; 
  if (!PL_unify_integer(blocks,buf.st_blocks))	PL_fail; 
  PL_succeed;
}

install_t
pl_stat_install()
{ 
  PL_register_foreign("stat", 10, pl_stat10, 0);
  PL_register_foreign("stat", 5, pl_stat5, 0);
  PL_register_foreign("lstat", 10, pl_lstat10, 0);
  PL_register_foreign("lstat", 5, pl_lstat5, 0);
}
